use std::path::PathBuf;

use anyhow::anyhow;
use serde::Deserialize;

#[derive(Deserialize, Clone)]
pub struct MemoryStorageConfig {
    pub max_size: usize,
    pub ttl: usize,
}

#[derive(Deserialize, Clone)]
pub struct DiskStorageConfig {
    pub path: PathBuf,
    pub ttl: usize,
    pub max_size: usize,
}

#[derive(Deserialize, Clone)]
pub struct MixedStorageConfig {
    pub path: PathBuf,
    pub ttl: usize,
    pub max_size: usize,
}

impl From<MixedStorageConfig> for MemoryStorageConfig {
    fn from(val: MixedStorageConfig) -> Self {
        MemoryStorageConfig {
            max_size: val.max_size,
            ttl: val.ttl,
        }
    }
}

impl From<MixedStorageConfig> for DiskStorageConfig {
    fn from(val: MixedStorageConfig) -> Self {
        DiskStorageConfig {
            path: val.path,
            ttl: val.ttl,
            max_size: val.max_size,
        }
    }
}

#[derive(Deserialize)]
pub enum StorageStrategyConfig {
    Memory(MemoryStorageConfig),
    Disk(DiskStorageConfig),
    Mixed(MixedStorageConfig),
}

#[derive(Deserialize)]
pub struct StorageConfig {
    pub strategy: StorageStrategyConfig,

    #[serde(with = "serde_regex")]
    pub regex: regex::Regex,
}

#[derive(Deserialize, Default)]
pub struct YAMLConfig {
    pub storages: Vec<StorageConfig>,
    pub secret_key: Option<String>,
    pub expose_api: Option<bool>,
}

pub async fn load() -> YAMLConfig {
    tokio::fs::read("config.yml")
        .await
        .map_err(|err| anyhow!(err))
        .and_then(|content| {
            serde_yaml::from_slice::<YAMLConfig>(&content).map_err(|err| anyhow!(err))
        })
        .unwrap_or_default()
}
