use std::{path::PathBuf, pin::Pin};

use crate::config::DiskStorageConfig;
use async_trait::async_trait;
use futures::TryStreamExt;
use tokio::io::AsyncWriteExt;
use tokio_stream::{Stream, StreamExt};
use tokio_util::{bytes::Bytes, io::ReaderStream};

use super::Storage;

pub struct DiskStorage {
    config: DiskStorageConfig,
}

impl DiskStorage {
    pub fn new(config: DiskStorageConfig) -> Self {
        Self { config }
    }

    pub async fn retrieve_all(
        &self,
    ) -> anyhow::Result<
        Vec<(
            String,
            Pin<Box<dyn Stream<Item = Result<Bytes, anyhow::Error>> + Send>>,
        )>,
    > {
        let mut saved_files = tokio::fs::read_dir(self.config.path.clone()).await?;
        let mut files = Vec::new();

        while let Some(file) = saved_files.next_entry().await? {
            let key = file.file_name().to_string_lossy().to_string();
            println!("{}", key);

            let stream = self.retrieve(key.clone()).await.unwrap();

            files.push((key, stream));
        }

        Ok(files)
    }
}

#[async_trait]
impl Storage for DiskStorage {
    async fn init(&mut self) -> anyhow::Result<()> {
        tokio::fs::create_dir_all(self.config.path.clone()).await?;

        Ok(())
    }

    async fn tick(&mut self) -> anyhow::Result<()> {
        Ok(())
    }

    async fn eligible(&self, _src: String) -> bool {
        true
    }

    async fn delete(&mut self, _key: String) -> anyhow::Result<()> {
        Ok(())
    }

    async fn retrieve(
        &self,
        key: String,
    ) -> Option<Pin<Box<dyn Stream<Item = Result<Bytes, anyhow::Error>> + Send>>> {
        let file_path = compute_file_path(self.config.path.clone(), key);

        match tokio::fs::File::open(file_path).await {
            Ok(file) => {
                let stream = ReaderStream::new(file).map_err(|_| anyhow::anyhow!(""));
                Some(Box::pin(stream))
            }
            Err(_) => None,
        }
    }

    async fn save(
        &mut self,
        key: String,
        mut stream: Pin<Box<dyn Stream<Item = Result<Bytes, anyhow::Error>> + Send>>,
    ) -> anyhow::Result<()> {
        let file_path = compute_file_path(self.config.path.clone(), key);
        let mut file = tokio::fs::OpenOptions::new()
            .write(true)
            .create(true)
            .open(file_path)
            .await?;

        while let Some(chunk) = stream.next().await {
            file.write_all(&chunk?).await?;
        }

        file.flush().await?;

        Ok(())
    }
}

fn compute_file_path(base: PathBuf, key: String) -> PathBuf {
    PathBuf::from_iter([base, key.into()])
}
